"""Logging utilities"""

import logging


class ConsecutiveDuplicateFilter(logging.Filter):
    def filter(self, record):
        current_log = (record.module, record.levelno, record.getMessage())
        if current_log != getattr(self, "last_log", None):
            self.last_log = current_log
            return True
        return False



from datetime import timedelta

import numpy as np

from westpa.tools import (
    WESTTool,
    WESTDataReader,
    IterRangeSelection,
)

# NumPy fixed time units (excludes nonlinear units 'Y and 'M')
TIME_UNITS = ('W', 'D', 'h', 'm', 's', 'ms', 'us', 'ns', 'ps', 'fs', 'as')


def _unit(delta):
    # Return the unit of a NumPy timedelta.
    words = str(delta.dtype).split('[')
    if len(words) == 1:
        return None
    return words[1][:-1]


def _str(delta):
    # Return a string representation of a NumPy timedelta.
    # Example: timedelta64(5380000, 'ps') -> '5.38 us'
    if _unit(delta) is None:
        return str(delta)
    for unit in TIME_UNITS:
        unit_delta = np.timedelta64(1, unit)
        try:
            if delta >= unit_delta:
                break
        except OverflowError:
            continue
    return f'{delta / unit_delta} {unit}'


def _delta(arg):
    # Construct a NumPy timedelta from an argument string.
    # Example: '100_ps' -> timedelta64(100, 'ps')
    try:
        value, unit = arg.split('_')
    except ValueError:
        raise ValueError('must be formatted as <value>_<unit>')
    if unit not in TIME_UNITS + ('μs',):  # accept either μs or us for microsecond
        raise ValueError(f'{unit!r} is not a recognized time unit')
    return np.timedelta64(int(value), unit)


class WTimings(WESTTool):
    prog = 'w_timings'
    description = 'Print timing information for a WESTPA simulation.'

    def __init__(self):
        super().__init__()
        self.data_reader = WESTDataReader()
        self.iter_range = IterRangeSelection(self.data_reader)
        self.tau = None

    def go(self):
        start = self.iter_range.iter_start - 1
        stop = self.iter_range.iter_stop - 1
        with self.data_reader:
            iter_summaries = self.data_reader.we_h5file['summary'][start:stop]

        walltime = iter_summaries['walltime'].sum()
        cputime = iter_summaries['cputime'].sum()
        n_particles = iter_summaries['n_particles'].sum()
        n_iters = len(iter_summaries)

        width = 26
        print('Iterations:'.ljust(width), n_iters)
        print('Total segments:'.ljust(width), n_particles)
        print('Wall-clock time:'.ljust(width), timedelta(seconds=walltime))
        if not np.isclose(cputime, 0):  # only print CPU time if it was recorded
            print('Total CPU time:'.ljust(width), timedelta(seconds=cputime))
        if self.tau is not None:
            print('Maximum trajectory length:', _str(n_iters * self.tau))
            print('Aggregate simulation time:', _str(n_particles * self.tau))

    def add_args(self, parser):
        self.data_reader.add_args(parser)
        self.iter_range.add_args(parser)

        unit_list = ', '.join(map(repr, TIME_UNITS[-1:0:-1])) + f', or {TIME_UNITS[0]!r}'
        parser.add_argument(
            '-t',
            '--tau',
            help=(
                'WE resampling interval (format: <value>_<unit>, where <value> '
                f'is a positive integer and <unit> is {unit_list}).'
            ),
        )

    def process_args(self, args):
        self.data_reader.process_args(args)
        with self.data_reader:
            self.iter_range.process_args(args)

        if args.tau is not None:
            try:
                self.tau = _delta(args.tau)
            except (TypeError, ValueError) as e:
                self.parser.error(f'argument -t/--tau: {e}')


def entry_point():
    WTimings().main()


if __name__ == "__main__":
    entry_point()



import pytest
from filecmp import cmpfiles
from io import StringIO

import numpy as np
import pickle
from numpy.testing import assert_array_equal

import westpa
from westpa.core.propagators.executable import ExecutablePropagator
from westpa.core.propagators.loaders import (
    numpy_data_loader,
    pickle_data_loader,
    aux_data_loader,
    restart_loader,
    restart_writer,
    seglog_loader,
    seglog_writer,
    pcoord_loader,
    mdanalysis_trajectory_loader,
    mdtraj_trajectory_loader,
)
from westpa.core.segment import Segment


class Test_Executable:
    """Class to test the propagator executable."""

    def test_data_config(self, ref_executable):
        """Test if the config is initialized correctly, where the executable propagator dataset options are set with the data manager options."""

        # Make the rc and executable read the config file.
        westpa.rc.read_config(filename='west_implicit.cfg')

        executable = ExecutablePropagator(rc=westpa.rc)

        check = {
            'pcoord': pcoord_loader,
            'displacement': numpy_data_loader,
            'trajectory': mdanalysis_trajectory_loader,
        }

        for dsname, loader in check.items():
            assert dsname in executable.data_info
            assert executable.data_info[dsname]['loader'] == loader

    def test_legacy_data_config(self, ref_executable):
        """Test if the dataset config is initialized correctly using the legacy part, where propagator datasets have to be specified twice."""

        # Make the rc and executable read the config file.
        westpa.rc.read_config(filename='west.cfg')

        executable = ExecutablePropagator(rc=westpa.rc)

        check = {
            'pcoord': pcoord_loader,
            'displacement': aux_data_loader,
            'trajectory': mdtraj_trajectory_loader,
        }

        for dsname, loader in check.items():
            assert dsname in executable.data_info
            assert executable.data_info[dsname]['loader'] == loader


class Test_Loaders:
    """Class to test if numpy_data_loader and pickle_data_loader are able to successfully add data into a dummy segment object."""

    def test_numpy_loader(self, ref_idtype):
        """Test if data loaded with numpy_data_loader is consistent."""

        test_segment = Segment()

        numpy_data_loader('test', self.correct_pkl, test_segment, False)

        with open(self.correct_pkl, 'rb') as f:
            ref_array = pickle.load(f)

        test_array = test_segment.data['test'][:]

        assert np.array_equal(test_array, ref_array)

    def test_pickle_loader(self, ref_idtype):
        """Test if data loaded with numpy_data_loader is consistent."""

        test_segment = Segment()

        pickle_data_loader('test', self.correct_pkl, test_segment, False)

        with open(self.correct_pkl, 'rb') as f:
            ref_array = pickle.load(f)

        test_array = test_segment.data['test'][:]

        assert np.array_equal(test_array, ref_array)

    def test_restart_loader_writer(self, nacl_restart_files):
        """Test if the restart file can be read, saved and reloaded correctly."""

        # Make a dummy segment and read/write the restart files
        test_segment = Segment()
        restart_loader('restart', self.return_dir, test_segment, False)
        restart_writer(self.write_dir, test_segment)

        # Do a shallow file comparison and make sure files tarred up and written out matches
        matches, mismatches, errors = cmpfiles(self.return_dir, self.write_dir, ['nacl.prmtop', 'nacl.ncrst'])
        assert sum([True if file in self.nacl_restart_files else False for file in matches]) == 2

    def test_seglog_loader_writer(self, nacl_restart_files):
        """Test if the log file can be saved and reloaded correctly."""

        # Make a dummy segment and read/write the seglog file
        test_segment = Segment()

        # Generate dummy log file
        dummy_text = 'abc\nlog\nend'
        self.test_file_path = self.test_dir / 'test.log'
        with open(self.test_file_path, 'w') as text_file:
            text_file.write(dummy_text)

        # Save the file into
        seglog_loader('log', self.test_file_path, test_segment, False)

        # Write the current file into self.write_dir
        seglog_writer(self.write_dir, test_segment)

        # Check to ensure contents are preserved
        with open(self.write_dir / 'seg.log', 'r') as text_file:
            assert text_file.read() == dummy_text

    def test_pcoord_loader_failures(self, ref_mab):
        test_segment = Segment()

        # Making test data
        rng = np.random.default_rng()
        c = rng.random(size=(11, 2), dtype=np.float32)
        io_file = StringIO()
        np.savetxt(io_file, c)

        with pytest.raises(AssertionError):
            pcoord_loader('test', c, test_segment, False)

        with pytest.raises(ValueError, match=r'incorrect shape \(11, 2\) \[expected \(2, 1\)\]'):
            io_file.seek(0)
            pcoord_loader('pcoord', io_file, test_segment, False)

    def test_pcoord_loader(self, ref_mab):
        test_segment = Segment()

        # Making test data
        rng = np.random.default_rng()
        c = rng.random(size=(2, 1), dtype=np.float32)
        io_file = StringIO()
        np.savetxt(io_file, c)

        io_file.seek(0)
        pcoord_loader('pcoord', io_file, test_segment, False)

        assert_array_equal(test_segment.pcoord, c)



import numpy as np
import os
from functools import cache

from mdtraj import Trajectory


def parseResidueAtoms(residue, map):
    """Parse all atoms from residue. Taken from OpenMM 8.2.0."""

    for atom in residue.findall('Atom'):
        name = atom.attrib['name']
        for id in atom.attrib:
            map[atom.attrib[id]] = name


@cache
def loadNameReplacementTables():
    """Load the list of atom and residue name replacements. Taken from OpenMM 8.2.0."""

    # importing things here because they're only used in this function
    try:
        from importlib.resources import files
    except ImportError:
        from importlib_resources import files

    import xml.etree.ElementTree as etree
    from copy import copy

    residueNameReplacements = {}
    atomNameReplacements = {}

    # This XML file is a to map all sorts of atom names/ residue names to the PDB 3.0 convention.
    tree = etree.parse(files('westpa') / 'data/pdbNames.xml')
    allResidues = {}
    proteinResidues = {}
    nucleicAcidResidues = {}
    for residue in tree.getroot().findall('Residue'):
        name = residue.attrib['name']
        if name == 'All':
            parseResidueAtoms(residue, allResidues)
        elif name == 'Protein':
            parseResidueAtoms(residue, proteinResidues)
        elif name == 'Nucleic':
            parseResidueAtoms(residue, nucleicAcidResidues)
    for atom in allResidues:
        proteinResidues[atom] = allResidues[atom]
        nucleicAcidResidues[atom] = allResidues[atom]
    for residue in tree.getroot().findall('Residue'):
        name = residue.attrib['name']
        for id in residue.attrib:
            if id == 'name' or id.startswith('alt'):
                residueNameReplacements[residue.attrib[id]] = name
        if 'type' not in residue.attrib:
            atoms = copy(allResidues)
        elif residue.attrib['type'] == 'Protein':
            atoms = copy(proteinResidues)
        elif residue.attrib['type'] == 'Nucleic':
            atoms = copy(nucleicAcidResidues)
        else:
            atoms = copy(allResidues)
        parseResidueAtoms(residue, atoms)
        atomNameReplacements[name] = atoms

    return residueNameReplacements, atomNameReplacements


def convert_mdanalysis_top_to_mdtraj(universe):
    """Convert a MDAnalysis Universe object's topology to a ``mdtraj.Topology`` object."""

    from mdtraj import Topology
    from mdtraj.core.element import get_by_symbol
    from MDAnalysis.exceptions import NoDataError

    top = Topology()  # Empty topology object
    residueNameReplacements, atomNameReplacements = loadNameReplacementTables()

    # Add in all the chains (called segments in MDAnalysis)
    for chain_segment in universe.segments:
        top.add_chain()

    all_chains = list(top.chains)

    # Add in all the residues
    for residue in universe.residues:
        try:
            resname = residueNameReplacements[residue.resname]
        except KeyError:
            resname = residue.resname

        top.add_residue(name=resname, chain=all_chains[residue.segindex], resSeq=residue.resid)

    all_residues = list(top.residues)

    # Add in all the atoms
    for atom, resid in zip(universe.atoms, universe.atoms.resindices):
        try:
            atomname = residueNameReplacements[atom.resname][atom.name]
        except (KeyError, TypeError):
            atomname = atom.name

        top.add_atom(name=atomname, element=get_by_symbol(atom.element), residue=all_residues[resid])

    all_atoms = list(top.atoms)

    # Add in all the bonds.  Depending on the topology type (e.g., pdb), there might not be bond information.
    try:
        for b_idx in universe.bonds._bix:
            top.add_bond(all_atoms[b_idx[0]], all_atoms[b_idx[1]])
    except NoDataError:
        top.create_standard_bonds()

    return top


class WESTTrajectory(Trajectory):
    """A subclass of ``mdtraj.Trajectory`` that contains the trajectory of atom coordinates with
    pointers denoting the iteration number and segment index of each frame."""

    def __init__(
        self,
        coordinates,
        topology=None,
        time=None,
        iter_labels=None,
        seg_labels=None,
        pcoords=None,
        parent_ids=None,
        unitcell_lengths=None,
        unitcell_angles=None,
    ):
        if isinstance(coordinates, Trajectory):
            xyz = coordinates.xyz
            topology = coordinates.topology if topology is None else topology
            time = coordinates.time if time is None else time
            unitcell_lengths = coordinates.unitcell_lengths if unitcell_lengths is None else unitcell_lengths
            unitcell_angles = coordinates.unitcell_angles if unitcell_angles is None else unitcell_angles
        else:
            xyz = coordinates

        super(WESTTrajectory, self).__init__(xyz, topology, time, unitcell_lengths, unitcell_angles)
        self._shape = None
        self.iter_labels = iter_labels
        self.seg_labels = seg_labels
        self.pcoords = pcoords
        self.parent_ids = parent_ids

    def _string_summary_basic(self):
        """Basic summary of WESTTrajectory in string form."""

        unitcell_str = 'and unitcells' if self._have_unitcell else 'without unitcells'
        value = "%s with %d frames, %d atoms, %d residues, %s" % (
            self.__class__.__name__,
            self.n_frames,
            self.n_atoms,
            self.n_residues,
            unitcell_str,
        )
        return value

    def _check_labels(self, value):
        if value is None:
            value = 0
        elif isinstance(value, list):
            value = np.array(value)

        if np.isscalar(value):
            value = np.array([value] * self.n_frames, dtype=int)
        elif value.shape != (self.n_frames,):
            raise ValueError('Wrong shape. Got %s, should be %s' % (value.shape, (self.n_frames,)))

        return value

    def _check_pcoords(self, value):
        if value is None:
            value = 0.0
        elif isinstance(value, list):
            value = np.array(value)

        if np.isscalar(value):
            value = np.array([(value,)] * self.n_frames, dtype=float)

        if value.ndim == 1:
            value = np.tile(value, (self.n_frames, 1))
        elif value.ndim != 2:
            raise ValueError('pcoords must be a 2-D array')

        elif value.shape[0] != self.n_frames:
            raise ValueError('Wrong length. Got %s, should be %s' % (value.shape[0], self.n_frames))

        return value

    def iter_label_values(self):
        visited_ids = []

        for i in self.iter_labels:
            if i in visited_ids:
                continue
            visited_ids.append(i)
            yield i

    def seg_label_values(self, iteration=None):
        seg_labels = self.seg_labels[self.iter_labels == iteration]
        visited_ids = []

        for j in seg_labels:
            if j in visited_ids:
                continue
            visited_ids.append(j)
            yield j

    @property
    def label_values(self):
        for i in self.iter_label_values():
            for j in self.seg_label_values(i):
                yield i, j

    def _iter_blocks(self):
        for i, j in self.label_values:
            IandJ = np.logical_and(self.iter_labels == i, self.seg_labels == j)
            yield i, j, IandJ

    @property
    def iter_labels(self):
        """Iteration index corresponding to each frame

        Returns
        -------
        time : np.ndarray, shape=(n_frames,)
            The iteration index corresponding to each frame
        """

        return self._iters

    @iter_labels.setter
    def iter_labels(self, value):
        """Set the iteration index corresponding to each frame"""

        self._iters = self._check_labels(value)
        self._shape = None

    @property
    def seg_labels(self):
        """
        Segment index corresponding to each frame

        Returns
        -------
        time : np.ndarray, shape=(n_frames,)
            The segment index corresponding to each frame
        """

        return self._segs

    @seg_labels.setter
    def seg_labels(self, value):
        """Set the segment index corresponding to each frame"""

        self._segs = self._check_labels(value)
        self._shape = None

    @property
    def pcoords(self):
        return self._pcoords

    @pcoords.setter
    def pcoords(self, value):
        self._pcoords = self._check_pcoords(value)

    @property
    def parent_ids(self):
        return self._parent_ids

    @parent_ids.setter
    def parent_ids(self, value):
        self._parent_ids = self._check_labels(value)

    def join(self, other, check_topology=True, discard_overlapping_frames=False):
        """
        Join two ``Trajectory``s. This overrides ``mdtraj.Trajectory.join``
        so that it also handles WESTPA pointers.
        ``mdtraj.Trajectory.join``'s documentation for more details.
        """

        if isinstance(other, Trajectory):
            other = [other]

        new_traj = super(WESTTrajectory, self).join(
            other, check_topology=check_topology, discard_overlapping_frames=discard_overlapping_frames
        )

        trajectories = [self] + other
        if discard_overlapping_frames:
            for i in range(len(trajectories) - 1):
                x0 = trajectories[i].xyz[-1]
                x1 = trajectories[i + 1].xyz[0]

                if np.all(np.abs(x1 - x0) < 2e-3):
                    trajectories[i] = trajectories[i][:-1]

        iter_labels = []
        seg_labels = []
        parent_ids = []
        pshape = self.pcoords.shape
        pcoords = []

        for t in trajectories:
            if hasattr(t, "iter_labels"):
                iters = t.iter_labels
            else:
                iters = np.zeros(len(t)) - 1  # default iter label: -1

            iter_labels.append(iters)

            if hasattr(t, "seg_labels"):
                segs = t.seg_labels
            else:
                segs = np.zeros(len(t)) - 1  # default seg label: -1

            seg_labels.append(segs)

            if hasattr(t, "parent_ids"):
                pids = t.parent_ids
            else:
                pids = np.zeros(len(t)) - 1  # default parent_id: -1

            parent_ids.append(pids)

            if hasattr(t, "pcoords"):
                p = t.pcoords
            else:
                p = np.zeros((len(t), pshape[-1]), dtype=float)  # default pcoord: 0.0

            pcoords.append(p)

        iter_labels = np.concatenate(iter_labels)
        seg_labels = np.concatenate(seg_labels)
        parent_ids = np.concatenate(parent_ids)
        pcoords = np.concatenate(pcoords)

        new_westpa_traj = WESTTrajectory(
            new_traj, iter_labels=iter_labels, seg_labels=seg_labels, pcoords=pcoords, parent_ids=parent_ids
        )

        return new_westpa_traj

    def slice(self, key, copy=True):
        """
        Slice the ``Trajectory``. This overrides ``mdtraj.Trajectory.slice``
        so that it also handles WESTPA pointers. Please see
        ``mdtraj.Trajectory.slice``'s documentation for more details.
        """

        if isinstance(key, tuple):
            if self._shape is None:
                uniq_iters = np.unique(self.iter_labels)
                max_iter = uniq_iters.max()
                max_seg = self.seg_labels.max()
                max_n_trajs = 0
                for _, _, block in self._iter_blocks():
                    n_trajs = block.sum()
                    if n_trajs > max_n_trajs:
                        max_n_trajs = n_trajs

                self._shape = (max_iter, max_seg, max_n_trajs)
            else:
                max_iter, max_seg, max_n_trajs = self._shape

            M = np.full((max_iter + 1, max_seg + 1, max_n_trajs), -1, dtype=int)
            all_traj_indices = np.arange(self.n_frames, dtype=int)
            for i, j, block in self._iter_blocks():
                traj_indices = all_traj_indices[block]

                for k, traj_idx in enumerate(traj_indices):
                    M[i, j, k] = traj_idx

            selected_indices = M[key].flatten()
            if np.isscalar(selected_indices):
                selected_indices = np.array([selected_indices])
            key = selected_indices[selected_indices != -1]

        iters = self.iter_labels[key]
        segs = self.seg_labels[key]
        pcoords = self.pcoords[key, :]
        parent_ids = self.parent_ids[key]

        traj = super(WESTTrajectory, self).slice(key, copy)
        traj.iter_labels = iters
        traj.seg_labels = segs
        traj.pcoords = pcoords
        traj.parent_ids = parent_ids

        return traj


def get_extension(filename):
    """A function to get the format extension of a file."""

    base, extension = os.path.splitext(filename)

    # Return the other part of the extension as well if it's a gzip.
    if extension == '.gz':
        return os.path.splitext(base)[1] + extension

    return extension


def find_top_traj_file(folder, eligible_top, eligible_traj):
    """
    A general (reusable) function for identifying and returning the appropriate
    file names in ``folder`` which are toplogy and trajectory. Useful when writing custom loaders.
    Note that it's possible that the topology_file and trajectory_file are identical.

    Parameters
    ----------
    folder : str or os.Pathlike
        A string or Pathlike to the folder to search.

    eligible_top : list of strings
        A list of accepted topology file extensions.

    eligible_traj : list of strings
        A list of accepted trajectory file extensions.


    Returns
    -------
    top_file : str
        Path to topology file

    traj_file : str
        Path to trajectory file
    """

    # Setting up the return variables
    top_file = traj_file = None

    # Extract a list of all files, ignoring hidden files that start with a '.'
    file_list = [f_name for f_name in os.listdir(folder) if not f_name.startswith('.')]

    for filename in file_list:
        filepath = os.path.join(folder, filename)
        if not os.path.isfile(filepath):
            continue

        ext = get_extension(filename).lower()
        # Catching trajectory formats that can be topology and trajectories at the same time.
        # Only activates when there is a single file in the folder.
        if len(file_list) < 2 and ext in eligible_top and ext in eligible_traj:
            top_file = filename
            traj_file = filename

        # Assuming topology file is copied first.
        if ext in eligible_top and top_file is None:
            top_file = filename
        elif ext in eligible_traj and traj_file is None:
            traj_file = filename

        if top_file is not None and traj_file is not None:
            break

    if traj_file is None:
        raise ValueError('trajectory file not found')

    traj_file = os.path.join(folder, traj_file)

    if top_file is not None:
        top_file = os.path.join(folder, top_file)

    return top_file, traj_file


@cache
def mdtraj_supported_extensions():
    from mdtraj import FormatRegistry, formats as mdformats
    from mdtraj.core.trajectory import _TOPOLOGY_EXTS

    FormatRegistry.loaders['.rst'] = mdformats.amberrst.load_restrt
    FormatRegistry.fileobjects['.rst'] = mdformats.AmberRestartFile
    FormatRegistry.loaders['.ncrst'] = mdformats.amberrst.load_ncrestrt
    FormatRegistry.fileobjects['.ncrst'] = mdformats.AmberRestartFile

    TRAJECTORY_EXTS = list(FormatRegistry.loaders.keys())
    TOPOLOGY_EXTS = list(_TOPOLOGY_EXTS)

    for ext in [".h5", ".hdf5", ".lh5"]:
        TOPOLOGY_EXTS.remove(ext)

    return TOPOLOGY_EXTS, TRAJECTORY_EXTS


@cache
def mdanalysis_supported_extensions():
    import MDAnalysis as mda

    TRAJECTORY_EXTS = [reader.format if isinstance(reader.format, list) else [reader.format] for reader in mda._READERS.values()]
    TRAJECTORY_EXTS = list(set(f'.{ext.lower()}' for ilist in TRAJECTORY_EXTS for ext in ilist))

    TOPOLOGY_EXTS = [parser.format if isinstance(parser.format, list) else [parser.format] for parser in mda._PARSERS.values()]
    TOPOLOGY_EXTS = list(set(f'.{ext.lower()}' for ilist in TOPOLOGY_EXTS for ext in ilist))

    return TOPOLOGY_EXTS, TRAJECTORY_EXTS


def load_mdtraj(folder):
    """
    Load trajectory from ``folder`` using ``mdtraj`` and return a ``mdtraj.Trajectory``
    object. The folder should contain a trajectory and a topology file (with a recognizable
    extension) that is supported by ``mdtraj``. The topology file is optional if the
    trajectory file contains topology data (e.g., HDF5 format).
    """

    from mdtraj import load as load_traj

    TOPOLOGY_EXTS, TRAJECTORY_EXTS = mdtraj_supported_extensions()

    top_file, traj_file = find_top_traj_file(folder, TOPOLOGY_EXTS, TRAJECTORY_EXTS)

    # MDTraj likes the (optional) topology part to be provided within a dictionary
    traj = load_traj(traj_file, **{'top': top_file})

    return traj


def load_netcdf(folder):
    """
    Load netcdf file from ``folder`` using ``scipy.io`` and return a ``mdtraj.Trajectory``
    object. The folder should contain a Amber trajectory file with extensions `.nc` or `.ncdf`.

    Note coordinates and box lengths are all divided by 10 to change from Angstroms to nanometers.
    """

    from scipy.io import netcdf_file

    _, traj_file = find_top_traj_file(folder, [], ['.nc', '.ncdf', '.ncrst'])

    # Extracting these datasets
    datasets = {'coordinates': None, 'cell_lengths': None, 'cell_angles': None, 'time': None}
    convert = ['coordinates', 'cell_lengths']  # Length-based datasets that need to be converted from Å to nm
    optional = ['cell_lengths', 'cell_angles']

    with netcdf_file(traj_file) as rootgrp:
        for key, val in datasets.items():
            if key in optional:
                pass
            elif key in convert and key in rootgrp.variables:
                datasets[key] = rootgrp.variables[key][()].copy() / 10  # From Å to nm
            else:
                datasets[key] = rootgrp.variables[key][()].copy()  # noqa: F841

    map_dataset = {
        'coordinates': datasets['coordinates'],
        'unitcell_lengths': datasets['cell_lengths'],
        'unitcell_angles': datasets['cell_angles'],
        'time': datasets['time'],
    }

    return WESTTrajectory(**map_dataset)


def load_mdanalysis(folder):
    """
    Load a file from ``folder`` using ``MDAnalysis`` and return a ``mdtraj.Trajectory``
    object. The folder should contain a trajectory and a topology file (with a recognizable
    extension) that is supported by ``MDAnalysis``. The topology file is optional if the
    trajectory file contains topology data (e.g., H5MD format).

    Note coordinates and box lengths are all divided by 10 to change from Angstroms to nanometers.
    """

    import MDAnalysis as mda

    TOPOLOGY_EXTS, TRAJECTORY_EXTS = mdanalysis_supported_extensions()

    top_file, traj_file = find_top_traj_file(folder, TOPOLOGY_EXTS, TRAJECTORY_EXTS)

    u = mda.Universe(top_file, traj_file)

    tot_frames = len(u.trajectory)
    coords = np.zeros((tot_frames, len(u.atoms), 3))
    time = np.zeros((tot_frames))

    # Periodic Boundary Conditions
    periodic = u.trajectory.periodic
    cell_lengths = np.zeros((tot_frames, 3)) if periodic else None
    cell_angles = np.zeros((tot_frames, 3)) if periodic else None

    for iframe, frame in enumerate(u.trajectory):
        coords[iframe] = frame._pos
        time[iframe] = frame.time

        if periodic:
            cell_lengths[iframe] = frame.dimensions[:3]
            cell_angles[iframe] = frame.dimensions[3:]

    # Length-based datasets that need to be converted
    convert = [coords, cell_lengths] if periodic else [coords]
    for dset in convert:
        dset = mda.units.convert(dset, 'angstrom', 'nanometer')

    traj = WESTTrajectory(coordinates=coords, unitcell_lengths=cell_lengths, unitcell_angles=cell_angles, time=time)

    return traj



import argparse
import os
from unittest import TestCase
from unittest.mock import MagicMock
import tempfile
import pytest

import numpy as np

import westpa
from westpa.core.binning.assign import RectilinearBinMapper
from westpa.core.binning.binless_manager import BinlessSimManager
from westpa.core.binning.mab_manager import MABSimManager
from westpa.core.segment import Segment
from westpa.core.states import BasisState
from westpa.core.sim_manager import PropagationError


def dummy_callback_one(self):
    pass


class TestSimManager(TestCase):
    def setUp(self):
        parser = argparse.ArgumentParser()
        westpa.rc.add_args(parser)

        here = os.path.dirname(__file__)
        os.environ['WEST_SIM_ROOT'] = os.path.join(here, 'refs', 'odld')

        config_file_name = os.path.join(here, 'refs', 'odld', 'west.cfg')
        args = parser.parse_args(['-r={}'.format(config_file_name)])
        westpa.rc.process_args(args)
        self.sim_manager = westpa.rc.get_sim_manager()
        self.test_dir = tempfile.mkdtemp()
        self.hdf5 = os.path.join(self.test_dir, "west.h5")
        self.basis_states = [BasisState(label="label", probability=1.0)]
        self.segments = [self.segment(0.0, 1.5, weight=0.125) for _i in range(4)] + [
            self.segment(1.5, 0.5, weight=0.125) for _i in range(4)
        ]
        self.sim_manager.we_driver.new_iteration()
        self.sim_manager.we_driver.assign(self.segments)
        self.sim_manager.we_driver.construct_next()
        self.sim_manager.segments = {segment.seg_id: segment for segment in self.segments}
        self.sim_manager.incomplete_segments = self.sim_manager.segments
        self.sim_manager.current_iter_istates = self.sim_manager.segments
        self.sim_manager.completed_segments = self.sim_manager.segments
        self.sim_manager.report_bin_statistics = MagicMock(return_value=True)

        self.work_manager = westpa.rc.get_work_manager()
        self.work_manager.running = True

        data = self.sim_manager.we_driver.rc.get_data_manager()
        data.we_h5filename = self.hdf5
        data.prepare_backing()
        data.create_ibstate_group([])
        data.create_initial_states(1)
        data.save_target_states([])
        data.update_segments = MagicMock(return_value=None)

        n_iter = 0
        it_name = data.iter_group_name(n_iter)
        for group in ["seg_index", "parents", "ibstates", "pcoord"]:
            data.we_h5file.create_group(it_name + "/" + group)
        data.get_new_weight_data = MagicMock(return_value=None)
        data.get_segments = MagicMock(return_value=self.segments)
        self.sim_manager.we_driver.rc.get_data_manager = MagicMock(return_value=data)
        self.sim_manager.n_iter = n_iter

    def tearDown(self):
        westpa.rc._sim_manager = None
        westpa.rc._system = None
        westpa.rc._data_manager = None
        del os.environ['WEST_SIM_ROOT']

    def dummy_callback_one(self):
        system = self.sim_manager.system
        bounds = [0.0, 1.0, 2.0, 3.0]
        system.bin_mapper = RectilinearBinMapper([bounds])

    def dummy_callback_two(self):
        system = self.sim_manager.system
        bounds = [0.0, 1.0, 2.0, 5.0]
        system.bin_mapper = RectilinearBinMapper([bounds])

    def segment(self, init_pcoord, final_pcoord, weight=1.0):
        segment = Segment(n_iter=1, seg_id=1123, pcoord=self.sim_manager.system.new_pcoord_array(), weight=weight)
        segment.pcoord[0] = init_pcoord
        segment.pcoord[1] = final_pcoord
        segment.parent_id = 1
        return segment

    def test_sim_manager(self):
        self.assertEqual(self.sim_manager.n_propagated, 0)
        self.assertEqual(len(self.sim_manager._callback_table), 0)

    def test_register_callback(self):
        hook = self.sim_manager.prepare_new_iteration

        self.sim_manager.register_callback(hook, self.dummy_callback_one, 3)
        self.sim_manager.register_callback(hook, self.dummy_callback_two, 0)
        self.sim_manager.register_callback(
            hook, dummy_callback_one, 3
        )  # Same name and priority, but different function, should be added
        self.sim_manager.register_callback(
            hook, self.dummy_callback_one, 2
        )  # Duplicate should never be added, even with different priority
        self.assertTrue(hook in self.sim_manager._callback_table)

        callbacks = self.sim_manager._callback_table.get(hook, [])

        assert len(callbacks) == 3  # Make sure only 3 added.

        self.assertTrue((3, self.dummy_callback_one.__name__, self.dummy_callback_one) in callbacks)  # noqa
        self.assertTrue((0, self.dummy_callback_two.__name__, self.dummy_callback_two) in callbacks)  # noqa
        self.assertTrue((3, dummy_callback_one.__name__, dummy_callback_one) in callbacks)  # noqa

    def test_invoke_callback(self):
        hook = self.sim_manager.prepare_new_iteration

        self.sim_manager.register_callback(hook, self.dummy_callback_one, 3)
        self.sim_manager.register_callback(hook, self.dummy_callback_two, 0)

        self.sim_manager.invoke_callbacks(hook)

        system = self.sim_manager.system
        self.assertTrue(np.all(system.bin_mapper.boundaries == np.array([0.0, 1.0, 2.0, 3.0])))  # noqa

    def test_process_config(self):
        self.sim_manager.process_config()
        self.assertTrue(self.sim_manager.do_gen_istates)
        self.assertEqual(self.sim_manager.propagator_block_size, 10000)
        self.assertFalse(self.sim_manager.save_transition_matrices)
        self.assertEqual(self.sim_manager.max_run_walltime, 10800)
        self.assertEqual(self.sim_manager.max_total_iterations, 100)

    def test_load_plugins(self):
        self.sim_manager.load_plugins()

    def test_report_bin_statistics(self):
        self.sim_manager.report_bin_statistics([0.0, 1.0, 2.0, 5.0])

    def test_get_bstate_pcoords(self):
        self.sim_manager.get_bstate_pcoords(self.basis_states)

    def test_report_basis_states(self):
        self.sim_manager.report_basis_states(self.basis_states)

    def test_report_target_states(self):
        self.sim_manager.report_target_states(self.basis_states)

    @pytest.mark.skip(reason="Cannot currently test WESimManager.initialize_simulation()")
    def test_initialize_simulation(self):
        # TODO: determine how to test self.simulation_manager.initialize_simulation()
        pass

    def test_prepare_iteration(self):
        self.sim_manager.prepare_new_iteration()
        self.sim_manager.prepare_iteration()

    def test_finalize_iteration(self):
        self.sim_manager.finalize_iteration()

    def test_get_istate_futures(self):
        self.sim_manager.get_istate_futures()

    def test_propagate(self):
        westpa.core.states.pare_basis_initial_states = MagicMock(return_value=([], []))
        self.sim_manager.propagate

    def test_save_bin_data(self):
        self.sim_manager.save_bin_data()

    def test_check_propagation(self):
        self.assertRaises(PropagationError, self.sim_manager.check_propagation)

    def test_run_we(self):
        self.sim_manager.run_we()

    def test_run(self):
        self.sim_manager.run()

    def test_prepare_run(self):
        self.sim_manager.prepare_run()

    def test_finalize_run(self):
        self.sim_manager.finalize_run()

    def test_pre_propagation(self):
        self.sim_manager.pre_propagation()

    def test_post_propagation(self):
        self.sim_manager.post_propagation()

    def test_pre_we(self):
        self.sim_manager.pre_we()

    def test_post_we(self):
        self.sim_manager.post_we()


class TestMABSimManager(TestSimManager):
    def setUp(self):
        parser = argparse.ArgumentParser()
        westpa.rc.add_args(parser)

        here = os.path.dirname(__file__)
        os.environ['WEST_SIM_ROOT'] = os.path.join(here, 'refs', 'odld')

        config_file_name = os.path.join(here, 'refs', 'odld', 'west_mab.cfg')
        args = parser.parse_args(['-r={}'.format(config_file_name)])
        westpa.rc.process_args(args)
        self.sim_manager = MABSimManager()
        self.test_dir = tempfile.mkdtemp()
        self.hdf5 = os.path.join(self.test_dir, "west.h5")
        self.basis_states = [BasisState(label="label", probability=1.0)]
        self.segments = [self.segment(0.0, 1.5, weight=0.125) for _i in range(4)] + [
            self.segment(1.5, 0.5, weight=0.125) for _i in range(4)
        ]
        self.sim_manager.we_driver.new_iteration()
        self.sim_manager.we_driver.assign(self.segments)
        self.sim_manager.we_driver.construct_next()
        self.sim_manager.segments = {segment.seg_id: segment for segment in self.segments}
        self.sim_manager.incomplete_segments = self.sim_manager.segments
        self.sim_manager.current_iter_istates = self.sim_manager.segments
        self.sim_manager.completed_segments = self.sim_manager.segments
        self.sim_manager.report_bin_statistics = MagicMock(return_value=True)

        data = self.sim_manager.we_driver.rc.get_data_manager()
        data.we_h5filename = self.hdf5
        data.prepare_backing()
        data.create_ibstate_group([])
        data.create_initial_states(1)
        data.save_target_states([])
        data.update_segments = MagicMock(return_value=None)

        n_iter = 0
        it_name = data.iter_group_name(n_iter)
        for group in ["seg_index", "parents", "ibstates", "pcoord"]:
            data.we_h5file.create_group(it_name + "/" + group)
        data.get_new_weight_data = MagicMock(return_value=None)
        data.get_segments = MagicMock(return_value=self.segments)
        self.sim_manager.we_driver.rc.get_data_manager = MagicMock(return_value=data)
        self.sim_manager.n_iter = n_iter


class TestBinlessSimManager(TestSimManager):
    def setUp(self):
        parser = argparse.ArgumentParser()
        westpa.rc.add_args(parser)

        here = os.path.dirname(__file__)
        os.environ['WEST_SIM_ROOT'] = os.path.join(here, 'refs', 'odld')

        config_file_name = os.path.join(here, 'refs', 'odld', 'west_binless.cfg')
        args = parser.parse_args(['-r={}'.format(config_file_name)])
        westpa.rc.process_args(args)
        self.sim_manager = BinlessSimManager()
        self.test_dir = tempfile.mkdtemp()
        self.hdf5 = os.path.join(self.test_dir, "west.h5")
        self.basis_states = [BasisState(label="label", probability=1.0)]
        self.segments = [self.segment(0.0, 1.5, weight=0.125) for _i in range(4)] + [
            self.segment(1.5, 0.5, weight=0.125) for _i in range(4)
        ]
        self.sim_manager.we_driver.new_iteration()
        self.sim_manager.we_driver.assign(self.segments)
        self.sim_manager.we_driver.construct_next()
        self.sim_manager.segments = {segment.seg_id: segment for segment in self.segments}
        self.sim_manager.incomplete_segments = self.sim_manager.segments
        self.sim_manager.current_iter_istates = self.sim_manager.segments
        self.sim_manager.completed_segments = self.sim_manager.segments
        self.sim_manager.report_bin_statistics = MagicMock(return_value=True)

        data = self.sim_manager.we_driver.rc.get_data_manager()
        data.we_h5filename = self.hdf5
        data.prepare_backing()
        data.create_ibstate_group([])
        data.create_initial_states(1)
        data.save_target_states([])
        data.update_segments = MagicMock(return_value=None)

        n_iter = 0
        it_name = data.iter_group_name(n_iter)
        for group in ["seg_index", "parents", "ibstates", "pcoord"]:
            data.we_h5file.create_group(it_name + "/" + group)
        data.get_new_weight_data = MagicMock(return_value=None)
        data.get_segments = MagicMock(return_value=self.segments)
        self.sim_manager.we_driver.rc.get_data_manager = MagicMock(return_value=data)
        self.sim_manager.n_iter = n_iter

    @pytest.mark.skip('Not configured')
    def test_run(self):
        self.sim_manager.run()



import pytest
from io import StringIO

import numpy as np

from westpa.core.binning.assign import RecursiveBinMapper
from westpa.tools.binning import mapper_from_expr, mapper_from_system, write_bin_info


class TestBinParsing:
    def test_mapper_from_expr(self):
        mapper = mapper_from_expr(r'[[-inf, 0, 15, np.inf]]')

        assert mapper.ndim == 1
        assert mapper.nbins == 3
        assert np.array_equal(mapper._boundaries, [np.array([-float('inf'), 0, 15, float('inf')], dtype=np.float32)])
        assert mapper.labels == ['[(-inf, 0.0)]', '[(0.0, 15.0)]', '[(15.0, inf)]']

    def test_mapper_from_expr_exceptions(self):
        with pytest.raises(NameError):
            mapper_from_expr('abc')

        with pytest.raises(TypeError):
            mapper_from_expr(int(3))

    def test_mapper_from_system(self, ref_mab):
        assert isinstance(mapper_from_system(), RecursiveBinMapper)


class TestBinningWriteout:
    def test_write_bin_info(self):
        out_io = StringIO()
        mapper = mapper_from_expr(r'[[-inf, 0, 15, np.inf]]')
        assignments = mapper.assign([[3], [3], [18]])
        weights = np.asarray([0.25, 0.25, 0.5])

        reference_output = '''3 segments
3 bins total, 1 targets, 2 (100.0%) occupied
Minimum probability by bin:     5.00000000000000000e-01
Maximum probability by bin:     5.00000000000000000e-01
Dynamic range (by bin):         -0 kT
Minimum probability by segment: 2.50000000000000000e-01
Maximum probability by segment: 5.00000000000000000e-01
Dynamic range (by segment):     0.693147 kT
Norm = 1, error in norm = 0 (0 epsilon)

 Index     Count    Total weight               Min seg weight             Max seg weight             Weight ratio    Label
     0         0    0.00000000000000000e+00    0.00000000000000000e+00    0.00000000000000000e+00         0.00000    [(-inf, 0.0)]
     1         2    5.00000000000000000e-01    2.50000000000000000e-01    2.50000000000000000e-01         1.00000    [(0.0, 15.0)]
     2         1    5.00000000000000000e-01    5.00000000000000000e-01    5.00000000000000000e-01         1.00000    [(15.0, inf)]
'''

        write_bin_info(mapper=mapper, assignments=assignments, weights=weights, n_target_states=1, outfile=out_io, detailed=True)
        assert out_io.getvalue() == reference_output



import os
import sys
import random
import logging
import threading
import traceback
import multiprocessing
from multiprocessing.queues import Empty

import westpa.work_managers as work_managers
from .core import WorkManager, WMFuture

log = logging.getLogger(__name__)

# Tasks are tuples ('task', task_id, fn, args, kwargs).
# Results are tuples (rtype, task_id, payload) where rtype is 'result' or 'exception' and payload is the return value
# or exception, respectively.

task_shutdown_sentinel = ('shutdown', None, None, (), {})
result_shutdown_sentinel = ('shutdown', None, None)


class ProcessWorkManager(WorkManager):
    '''A work manager using the ``multiprocessing`` module.

    Notes
    -----

    On MacOS, as of Python 3.8 the default start method for multiprocessing launching new processes was changed from fork to spawn.
    On Linux, as of Python 3.14, the default start method for multiprocessing launching new processes was changed from fork to spawn.
    In general, spawn is more robust and efficient, however it requires serializability of everything being passed to the child process.
    In contrast, fork is much less memory efficient, as it makes a full copy of everything in the parent process.
    However, it does not require picklability.

    So, on MacOS and Linux, the method for launching new processes is explicitly changed to fork from the (UNIX-specific) default of spawn.
    Unix should default to fork.

    See https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods and
    https://docs.python.org/3/library/multiprocessing.html#the-spawn-and-forkserver-start-methods for more details.
    '''

    @classmethod
    def from_environ(cls, wmenv=None):
        if wmenv is None:
            wmenv = work_managers.environment.default_env
        return cls(wmenv.get_val('n_workers', multiprocessing.cpu_count(), int))

    def __init__(self, n_workers=None, shutdown_timeout=1):
        super().__init__()

        try:
            if sys.platform in ['darwin', 'linux']:  # UNIX Platforms
                multiprocessing.set_start_method('fork')
                log.debug('setting multiprocessing start method to fork')
        except RuntimeError:
            log.debug('failed to set start method to fork')

        self.n_workers = n_workers or multiprocessing.cpu_count()
        self.workers = None
        self.task_queue = multiprocessing.Queue()
        self.result_queue = multiprocessing.Queue()
        self.receive_thread = None
        self.pending = None

        self.shutdown_received = threading.Event()
        self.shutdown_timeout = shutdown_timeout or 1

    def task_loop(self):
        # Close standard input, so we don't get SIGINT from ^C
        try:
            sys.stdin.close()
        except Exception as e:
            log.info("can't close stdin: {}".format(e))

        # (re)initialize random number generator in this process
        random.seed()

        while not self.shutdown_received.is_set():
            if not self.task_queue.empty():
                message, task_id, fn, args, kwargs = self.task_queue.get()[:5]

                if message == 'shutdown':
                    break
                try:
                    result = fn(*args, **kwargs)
                except BaseException as e:
                    result_tuple = ('exception', task_id, (e, traceback.format_exc()))
                else:
                    result_tuple = ('result', task_id, result)
                self.result_queue.put(result_tuple)

        log.debug('exiting task_loop')
        return

    def results_loop(self):
        while not self.shutdown_received.is_set():
            if not self.result_queue.empty():
                message, task_id, payload = self.result_queue.get()[:3]

                if message == 'shutdown':
                    break
                elif message == 'exception':
                    future = self.pending.pop(task_id)
                    future._set_exception(*payload)
                elif message == 'result':
                    future = self.pending.pop(task_id)
                    future._set_result(payload)
                else:
                    raise AssertionError('unknown message {!r}'.format((message, task_id, payload)))

        log.debug('exiting results_loop')
        return

    def submit(self, fn, args=None, kwargs=None):
        ft = WMFuture()
        log.debug('dispatching {!r}'.format(fn))
        self.pending[ft.task_id] = ft
        self.task_queue.put(('task', ft.task_id, fn, args or (), kwargs or {}))
        return ft

    def startup(self):
        from . import environment

        if not self.running:
            log.debug('starting up work manager {!r}'.format(self))
            self.running = True
            self.workers = [
                multiprocessing.Process(target=self.task_loop, name='worker-{:d}-{:x}'.format(i, id(self)))
                for i in range(self.n_workers)
            ]

            pi_name = '{}_PROCESS_INDEX'.format(environment.WMEnvironment.env_prefix)
            for iworker, worker in enumerate(self.workers):
                os.environ[pi_name] = str(iworker)
                worker.start()
            try:
                del os.environ[pi_name]
            except KeyError:
                pass

            self.pending = dict()

            self.receive_thread = threading.Thread(target=self.results_loop, name='receiver')
            self.receive_thread.daemon = True
            self.receive_thread.start()

    def _empty_queues(self):
        '''Empty self.task_queue and self.result_queue until queue is empty'''
        try:
            while True:
                self.task_queue.get_nowait()
        except (Empty, ValueError):
            log.debug('Empty task_queue')

        try:
            while True:
                self.result_queue.get_nowait()
        except (Empty, ValueError):
            log.debug('Empty result_queue')

    def shutdown(self):
        while self.running:
            log.debug('shutting down {!r}'.format(self))

            # Empty queues
            self._empty_queues()
            for _i in range(self.n_workers):
                self.task_queue.put_nowait(task_shutdown_sentinel)
            self.result_queue.put_nowait(result_shutdown_sentinel)

            # Signal shutdown Event to stop queue loops
            self.shutdown_received.set()
            self._empty_queues()

            # Terminating all workers
            for worker in self.workers:
                worker.join(self.shutdown_timeout)
                if worker.is_alive():
                    log.debug('sending SIGINT to worker process {:d}'.format(worker.pid))
                    worker.terminate()
                    worker.join(self.shutdown_timeout)
                    if worker.is_alive():
                        log.warning('sending SIGKILL to worker process {:d}'.format(worker.pid))
                        worker.kill()
                        worker.join()

                    log.debug('worker process {:d} terminated with code {:d}'.format(worker.pid, worker.exitcode))
                else:
                    log.debug('worker process {:d} terminated gracefully with code {:d}'.format(worker.pid, worker.exitcode))

                try:
                    worker.close()  # Release all resources
                except ValueError:
                    try:
                        if worker.is_alive():
                            log.debug('worker process {:d} could not be closed'.format(worker.pid))
                    except ValueError:
                        pass  # Already closed.

            self.running = False
            log.debug('Done shutting down the processes work manager')



import concurrent.futures as cf
import functools
import inspect
import operator
import os
from typing import Callable

try:
    import mdtraj
except ImportError:
    mdtraj = None
import numpy as np
from tqdm import tqdm

from westpa.analysis.core import Walker, Trace
from westpa.core.states import InitialState
from westpa.core.h5io import WESTIterationFile


class Trajectory:
    """A callable that returns the trajectory of a walker or trace.

    Parameters
    ----------
    fget : callable
        Function for retrieving a single trajectory segment. Must take a
        :class:`Walker` instance as its first argument and accept a boolean
        keyword argument `include_initpoint`. The function should return a
        sequence (e.g., a list or ndarray) representing the trajectory of
        the walker. If `include_initpoint` is True, the trajectory segment
        should include its initial point. Otherwise, the trajectory segment
        should exclude its initial point.
    fconcat : callable, optional
        Function for concatenating trajectory segments. Must take a sequence
        of trajectory segments as input and return their concatenation. The
        default concatenation function is :func:`concatenate`.

    """

    def __init__(self, fget=None, *, fconcat=None):
        if fget is None:
            return functools.partial(self.__init__, fconcat=fconcat)

        if 'include_initpoint' not in inspect.signature(fget).parameters:
            raise ValueError("'fget' must accept a parameter 'include_initpoint'")

        self._fget = fget
        self.fconcat = fconcat

        self._segment_collector = SegmentCollector(self)

    @property
    def segment_collector(self):
        """SegmentCollector: Segment retrieval manager."""
        return self._segment_collector

    @property
    def fget(self):
        """callable: Function for getting trajectory segments."""
        return self._fget

    @property
    def fconcat(self):
        """callable: Function for concatenating trajectory segments."""
        return self._fconcat

    @fconcat.setter
    def fconcat(self, value):
        if value is None:
            value = concatenate
        elif not isinstance(value, Callable):
            raise TypeError("'fconcat' must be a callable object")
        self._fconcat = value

    def __call__(self, obj, include_initpoint=True, **kwargs):
        if isinstance(obj, Walker):
            value = self.fget(obj, include_initpoint=include_initpoint, **kwargs)
            self._validate_segment(value)
            return value
        if isinstance(obj, Trace):
            initpoint_mask = np.full(len(obj), False)
            initpoint_mask[0] = include_initpoint
            segments = self.segment_collector.get_segments(obj, initpoint_mask, **kwargs)
            return self.fconcat(segments)
        raise TypeError('argument must be a Walker or Trace instance')

    def _validate_segment(self, value):
        if not hasattr(value, '__getitem__'):
            msg = f"{type(value).__name__!r} object can't be concatenated"
            raise TypeError(msg)


class SegmentCollector:
    """An object that manages the retrieval of trajectory segments.

    Parameters
    ----------
    trajectory : Trajectory
        The trajectory to which the segment collector is attached.
    use_threads : bool, default False
        Whether to use a pool of threads to retrieve trajectory segments
        asynchronously. Setting this parameter to True may be
        useful when segment retrieval is an I/O bound task.
    max_workers : int, optional
        Maximum number of threads to use. The default value is specified in the
        `ThreadPoolExecutor <https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor>`_
        documentation.
    show_progress : bool, default False
        Whether to show a progress bar when retrieving multiple segments.

    """

    def __init__(self, trajectory, use_threads=False, max_workers=None, show_progress=False):
        self.trajectory = trajectory
        self.use_threads = use_threads
        self.max_workers = max_workers
        self.show_progress = show_progress

    @property
    def trajectory(self):
        return self._trajectory

    @trajectory.setter
    def trajectory(self, value):
        if not isinstance(value, Trajectory):
            msg = f'trajectory must be an instance of {Trajectory}'
            raise TypeError(msg)
        self._trajectory = value

    @property
    def use_threads(self):
        return self._use_threads

    @use_threads.setter
    def use_threads(self, value):
        if not isinstance(value, bool):
            raise TypeError('use_threads must be True or False')
        self._use_threads = value

    @property
    def max_workers(self):
        return self._max_workers

    @max_workers.setter
    def max_workers(self, value):
        if value is None:
            self._max_workers = None
            return
        if value <= 0:
            raise ValueError('max_workers must be greater than 0')
        self._max_workers = value

    @property
    def show_progress(self):
        return self._show_progress

    @show_progress.setter
    def show_progress(self, value):
        if not isinstance(value, bool):
            raise ValueError('show_progress must be True or False')
        self._show_progress = value

    def get_segments(self, walkers, initpoint_mask=None, **kwargs):
        """Retrieve the trajectories of multiple walkers.

        Parameters
        ----------
        walkers : sequence of Walker
            The walkers for which to retrieve trajectories.
        initpoint_mask : sequence of bool, optional
            A Boolean mask indicating whether each trajectory segment should
            include (True) or exclude (False) its initial point. Default is
            all True.

        Returns
        -------
        list of sequences
            The trajectory of each walker.

        """
        if initpoint_mask is None:
            initpoint_mask = np.full(len(walkers), True)
        else:
            initpoint_mask = np.asarray(initpoint_mask, dtype=bool)

        get_segment = functools.partial(self.trajectory, **kwargs)

        tqdm_kwargs = dict(
            desc='Retrieving segments',
            disable=(not self.show_progress),
            position=0,
            total=len(walkers),
        )

        if self.use_threads:
            with cf.ThreadPoolExecutor(self.max_workers) as executor:
                future_to_key = {
                    executor.submit(get_segment, walker, include_initpoint=i): key
                    for key, (walker, i) in enumerate(zip(walkers, initpoint_mask))
                }
                futures = list(tqdm(cf.as_completed(future_to_key), **tqdm_kwargs))
                futures.sort(key=future_to_key.get)
                segments = (future.result() for future in futures)
        else:
            it = (get_segment(walker, include_initpoint=i) for walker, i in zip(walkers, initpoint_mask))
            segments = tqdm(it, **tqdm_kwargs)

        return list(segments)


class BasicMDTrajectory(Trajectory):
    """Trajectory reader for MD trajectories stored as in the
    `Basic Tutorial <https://github.com/westpa/westpa_tutorials/tree/main/tutorial7.1-basic-nacl>`_.

    Parameters
    ----------
    top : str or mdtraj.Topology, default 'bstate.pdb'
    traj_ext : str, default '.dcd'
    state_ext : str, default '.xml'
    sim_root : str, default '.'

    """

    def __init__(self, top='bstate.pdb', traj_ext='.dcd', state_ext='.xml', sim_root='.'):
        if mdtraj is None:
            raise ImportError('MDTraj must be installed to use the BasicMDTrajectory reader')

        self.top = top
        self.traj_ext = traj_ext
        self.state_ext = state_ext
        self.sim_root = sim_root

        def fget(walker, include_initpoint=True, atom_indices=None, sim_root=None):
            sim_root = sim_root or self.sim_root

            if isinstance(self.top, str):
                top = os.path.join(sim_root, 'common_files', self.top)
            else:
                top = self.top

            path = os.path.join(
                sim_root,
                'traj_segs',
                format(walker.iteration.number, '06d'),
                format(walker.index, '06d'),
                'seg' + self.traj_ext,
            )
            if top is not None:
                traj = mdtraj.load(path, top=top)
            else:
                traj = mdtraj.load(path)

            if include_initpoint:
                parent = walker.parent

                if isinstance(parent, InitialState):
                    if parent.istate_type == InitialState.ISTATE_TYPE_BASIS:
                        path = os.path.join(
                            sim_root,
                            'bstates',
                            parent.basis_state.auxref,
                        )
                    else:
                        path = os.path.join(
                            sim_root,
                            'istates',
                            str(parent.iter_created),
                            str(parent.state_id) + self.state_ext,
                        )
                else:
                    path = os.path.join(
                        sim_root,
                        'traj_segs',
                        format(walker.iteration.number - 1, '06d'),
                        format(parent.index, '06d'),
                        'seg' + self.state_ext,
                    )

                frame = mdtraj.load(path, top=traj.top)
                traj = frame.join(traj, check_topology=False)

            if atom_indices is not None:
                traj.atom_slice(atom_indices, inplace=True)

            return traj

        super().__init__(fget)

        self.segment_collector.use_threads = True
        self.segment_collector.show_progress = True


class HDF5MDTrajectory(Trajectory):
    """Trajectory reader for MD trajectories stored by the HDF5 framework."""

    def __init__(self):
        if mdtraj is None:
            raise ImportError('MDTraj must be installed to use the HDF5MDTrajectory reader')

        def fget(walker, include_initpoint=True, atom_indices=None):
            iteration = walker.iteration

            try:
                link = iteration.h5group['trajectories']
            except KeyError:
                msg = 'the HDF5 framework does not appear to have been used to store trajectories for this run'
                raise ValueError(msg)

            with WESTIterationFile(link.file.filename) as traj_file:
                traj = traj_file.read_as_traj(
                    iteration=iteration.number,
                    segment=walker.index,
                    atom_indices=atom_indices,
                )

            if include_initpoint:
                parent = walker.parent

                if isinstance(parent, InitialState):
                    if parent.istate_type == InitialState.ISTATE_TYPE_BASIS:
                        link = walker.run.h5file.get_iter_group(0)['trajectories']
                        with WESTIterationFile(link.file.filename) as traj_file:
                            frame = traj_file.read_as_traj(
                                iteration=0,
                                segment=parent.basis_state_id,
                                atom_indices=atom_indices,
                            )
                    elif parent.istate_type == InitialState.ISTATE_TYPE_GENERATED:
                        link = walker.run.h5file.get_iter_group(0)['trajectories']
                        istate_iter = -int(parent.iter_created)  # the conversion to int is because iter_created is uint
                        with WESTIterationFile(link.file.filename) as traj_file:
                            frame = traj_file.read_as_traj(
                                iteration=istate_iter,
                                segment=parent.state_id,
                                atom_indices=atom_indices,
                            )
                    else:
                        raise ValueError('unsupported initial state type: %d' % parent.istate_type)
                else:
                    frame = fget(parent, include_initpoint=False, atom_indices=atom_indices)[-1]
                traj = frame.join(traj, check_topology=False)

            return traj

        super().__init__(fget)

        self.segment_collector.use_threads = False
        self.segment_collector.show_progress = True


def concatenate(segments):
    """Return the concatenation of a sequence of trajectory segments.

    Parameters
    ----------
    segments : sequence of sequences
        A sequence of trajectory segments.

    Returns
    -------
    sequence
        The concatenation of `segments`.

    """
    if isinstance(segments[0], np.ndarray):
        return np.concatenate(segments)
    if mdtraj is not None and isinstance(segments[0], mdtraj.Trajectory):
        return segments[0].join(segments[1:], check_topology=False)
    return functools.reduce(operator.concat, segments)



import itertools
from pathlib import Path

import h5py
import numpy as np
import pandas as pd
import pytest

from westpa.analysis import Run
from westpa.analysis.core import Iteration, Walker
from westpa.core.binning import RectilinearBinMapper
from westpa.core.h5io import WESTPAH5File
from westpa.core.states import BasisState, InitialState, TargetState


class TestAnalysis:
    @pytest.fixture
    def h5filename(self) -> str:
        p = Path(__file__).parents[0] / 'refs' / 'west_ref.h5'
        return str(p)

    @pytest.fixture
    def run(self, h5filename) -> Run:
        run = Run.open(h5filename)
        yield run
        run.close()

    def test_open_close(self, h5filename):
        run_obj = Run.open(h5filename)
        assert not run_obj.closed
        run_obj.close()
        assert run_obj.closed

    def test_context_manager(self, h5filename):
        with Run.open(h5filename) as run_obj:
            assert not run_obj.closed
        assert run_obj.closed

    def test_num_iterations(self, run):
        assert run.num_iterations == 50
        assert len(run) == run.num_iterations  # Run.__len__

    def test_iteration(self, run):
        for n in range(1, run.num_iterations + 1):
            iteration = run.iteration(n)
            assert isinstance(iteration, Iteration)
            assert iteration in run
            assert iteration.run is run
            assert iteration.number == n

    def test_iterations(self, run):
        assert isinstance(run.iterations, list)
        assert run.iterations == list(run)  # Run.__iter__, Iteration.__eq__
        for iteration, n in zip(run.iterations, range(1, run.num_iterations + 1)):
            assert isinstance(iteration, Iteration)
            assert iteration in run
            assert iteration.run is run
            assert iteration.number == n

    def test_num_walkers(self, run):
        assert run.num_walkers == 9985
        assert sum(iteration.num_walkers for iteration in run) == run.num_walkers
        assert run.num_walkers == run.num_segments  # alias
        for iteration in run:
            assert iteration.num_walkers == iteration.num_segments  # alias

    def test_walker(self, run):
        iteration = run.iteration(1)
        for i in range(iteration.num_walkers):
            walker = iteration.walker(i)
            assert isinstance(walker, Walker)
            assert walker in run
            assert walker in iteration
            assert walker.run is run
            assert walker.iteration is iteration
            assert walker.index == i

    def test_walkers(self, run):
        for walker in run.walkers:
            assert isinstance(walker, Walker)
            assert walker in run
            assert walker.run is run

        assert len(list(run.walkers)) == run.num_walkers

        for iteration in run:
            assert list(iteration.walkers) == list(iteration)  # Iteration.__iter__, Walker.__eq__
            for i, walker in enumerate(iteration.walkers):
                assert isinstance(walker, Walker)
                assert walker in run
                assert walker in iteration
                assert walker.run is run
                assert walker.iteration is iteration
                assert walker.index == i
        assert sum(len(list(iteration.walkers)) for iteration in run) == run.num_walkers

    def test_parent(self, run):
        for walker in run.iteration(1):
            assert isinstance(walker.parent, InitialState)
        for walker in run.iteration(2):
            assert isinstance(walker.parent, Walker)
            assert walker.parent.iteration.number == 1

    def test_children(self, run):
        for walker in run.iteration(1):
            for child in walker.children:
                assert isinstance(child, Walker)
                assert child.iteration.number == 2

    def test_recycled(self, run):
        for walker in run.iteration(1):
            assert not walker.recycled

    def test_initial(self, run):
        for walker in run.iteration(1):
            assert walker.initial

    def test_recycled_walkers(self, run):
        for walker in run.recycled_walkers:
            assert walker.recycled
        assert len(list(run.recycled_walkers)) == 39
        for walker1, walker2 in zip(
            run.recycled_walkers,
            itertools.chain(*(iteration.recycled_walkers for iteration in run)),
        ):
            assert walker1 == walker2

    def test_initial_walkers(self, run):
        for walker in run.initial_walkers:
            assert walker.initial
            assert isinstance(walker.parent, InitialState)
        assert len(list(run.initial_walkers)) == 5
        for walker1, walker2 in zip(
            run.initial_walkers,
            itertools.chain(*(iteration.initial_walkers for iteration in run)),
        ):
            assert walker1 == walker2

    def test_auxiliary_data(self, run):
        for iteration in run:
            assert iteration.auxiliary_data is None
        for walker in run.iteration(1):
            assert walker.auxiliary_data == {}

    def test_basis_state_summaries(self, run):
        for iteration in run:
            summaries = iteration.basis_state_summaries
            assert isinstance(summaries, pd.DataFrame)
            assert list(summaries.axes[1]) == ['label', 'probability', 'auxref']

    def test_basis_state_pcoords(self, run):
        for iteration in run:
            assert iteration.basis_state_pcoords.ndim == 2

    def test_basis_states(self, run):
        for iteration in run:
            basis_states = iteration.basis_states
            assert isinstance(basis_states, list)
            assert all(isinstance(state, BasisState) for state in basis_states)

    def test_has_target_states(self, run):
        for iteration in run:
            assert iteration.has_target_states

    def test_target_state_summaries(self, run):
        for iteration in run:
            summaries = iteration.target_state_summaries
            assert isinstance(summaries, pd.DataFrame)
            assert list(summaries.axes[1]) == ['label']

    def test_target_state_pcoords(self, run):
        for iteration in run:
            assert iteration.target_state_pcoords.ndim == 2

    def test_target_states(self, run):
        for iteration in run:
            target_states = iteration.target_states
            assert isinstance(target_states, list)
            assert all(isinstance(state, TargetState) for state in target_states)

    def test_sink(self, run):
        for walker in run.recycled_walkers:
            assert walker.pcoords[0] not in walker.iteration.sink
            assert walker.pcoords[-1] in walker.iteration.sink

    def test_summary(self, run):
        fields = [
            'n_particles',
            'min_bin_prob',
            'max_bin_prob',
            'min_seg_prob',
            'max_seg_prob',
            'cputime',
            'walltime',
        ]
        summary = run.summary
        assert isinstance(summary, pd.DataFrame)
        assert len(summary) == len(run)
        assert all(summary.index == range(1, run.num_iterations + 1))
        assert list(summary.axes[1]) == fields
        for iteration in run:
            summary = iteration.summary
            assert isinstance(summary, pd.Series)
            assert list(summary.index) == fields
            assert summary.name == iteration.number

    def test_segment_summaries(self, run):
        fields = [
            'weight',
            'parent_id',
            'wtg_n_parents',
            'wtg_offset',
            'cputime',
            'walltime',
            'endpoint_type',
            'status',
        ]
        for iteration in run:
            summaries = iteration.segment_summaries
            assert isinstance(summaries, pd.DataFrame)
            assert len(summaries) == iteration.num_walkers
            assert all(summaries.index == range(iteration.num_walkers))
            assert list(summaries.axes[1]) == fields
        for walker in run.iteration(1):
            summary = walker.segment_summary
            assert isinstance(summary, pd.Series)
            assert list(summary.index) == fields

    def test_h5filename(self, run, h5filename):
        assert run.h5filename == h5filename

    def test_h5file(self, run):
        assert isinstance(run.h5file, WESTPAH5File)
        assert run.h5file.filename == run.h5filename

    def test_h5group(self, run):
        for iteration in run:
            assert isinstance(iteration.h5group, h5py.Group)
            assert iteration.h5group == run.h5file.get_iter_group(iteration.number)

    def test_prev_next(self, run):
        for iteration1, iteration2 in zip(run.iterations[:-1], run.iterations[1:]):
            assert iteration1.next == iteration2
            assert iteration2.prev == iteration1

    def test_pcoords(self, run):
        for iteration in run:
            assert iteration.pcoords.ndim == 3
            assert len(iteration.pcoords) == iteration.num_walkers
        pcoords = run.iteration(1).pcoords
        for walker in run.iteration(1):
            assert np.allclose(walker.pcoords, pcoords[walker.index])

    def test_weights(self, run):
        for iteration in run:
            assert iteration.weights.ndim == 1
            assert len(iteration.weights) == iteration.num_walkers
        weights = run.iteration(1).weights
        for walker in run.iteration(1):
            assert np.isclose(walker.weight, weights[walker.index])

    def test_binning(self, run):
        for iteration in run:
            bin_mapper = iteration.bin_mapper
            if bin_mapper is None:
                assert iteration.bin_target_counts is None
                assert iteration.num_bins == 0
            else:
                assert isinstance(bin_mapper, RectilinearBinMapper)
                assert len(iteration.bin_target_counts) == bin_mapper.nbins
                assert iteration.num_bins == bin_mapper.nbins

    def test_trace(self, run):
        walker = run.iterations[-1].walker(0)
        trace = walker.trace()
        assert len(trace) == walker.iteration.number
        assert all(isinstance(walker, Walker) for walker in trace)
        assert isinstance(trace.initial_state, InitialState)



import numpy as np

from westpa.core.h5io import WESTIterationFile
from westpa.core.segment import Segment


class Test_H5io:
    """Class to test the h5io module."""

    def test_write_segment(self, west_iteration_file):
        """Segment is written successfully with WestIterationFile when there are duplicate pointers (and not)."""

        segment = Segment(n_iter=3, seg_id=5, data=self.dummy_data)
        shape = self.dummy_data['iterh5/trajectory'][:].shape

        with WESTIterationFile(self.h5_iter_file_path, mode='a') as h5_iter_file:
            h5_iter_file.write_segment(segment)
            assert np.allclose(h5_iter_file.root['coordinates'][:], self.dummy_data['iterh5/trajectory'][:])

            self.dummy_data['iterh5/trajectory'][:] *= 2

            # Rewriting with new coordinates and see if it is written in correctly
            h5_iter_file.write_segment(segment)

            assert np.all(h5_iter_file.root['coordinates'][:].shape == shape), 'Extra frames added to per-iter HDF5 File'
            assert np.allclose(
                h5_iter_file.root['coordinates'][:], self.dummy_data['iterh5/trajectory'][:]
            ), 'Unable to overwrite trajectory coordinates'

            # Writing a new segment
            segment2 = Segment(n_iter=3, seg_id=6, data=self.dummy_data)
            h5_iter_file.write_segment(segment2)
            old_shape0 = shape[0]
            shape = (shape[0] * 2,) + shape[1:]

            assert np.all(
                h5_iter_file.root['coordinates'][:].shape == shape
            ), 'Extra frames not added to per-iter HDF5 File correctly'
            assert np.allclose(
                h5_iter_file.root['coordinates'][-old_shape0:], self.dummy_data['iterh5/trajectory'][:]
            ), 'Unable to write extra trajectory coordinates'

    def test_write_segment_under(self, west_iteration_file):
        """Segment is written successfully in WestIterationFile when there are not enough rows."""

        oneframe_data = {'iterh5/trajectory': self.dummy_data['iterh5/trajectory'][:1].copy()}
        self.dummy_data['iterh5/trajectory'][:] *= 2

        segment = Segment(n_iter=3, seg_id=5, data=self.dummy_data)
        segment_oneframe = Segment(n_iter=3, seg_id=5, data=oneframe_data)
        shape = self.dummy_data['iterh5/trajectory'][:].shape

        with WESTIterationFile(self.h5_iter_file_path, mode='a') as h5_iter_file:
            h5_iter_file.write_segment(segment_oneframe)
            assert np.allclose(h5_iter_file.root['coordinates'][:], oneframe_data['iterh5/trajectory'][:])
            print(self.dummy_data['iterh5/trajectory'][:])
            print(h5_iter_file.root['coordinates'][:])
            print(h5_iter_file.root['pointer'][:])
            # Rewriting with new coordinates and see if it is written in correctly
            h5_iter_file.write_segment(segment)

            assert np.all(h5_iter_file.root['coordinates'][:].shape == shape), 'Extra frames added to per-iter HDF5 File'
            assert np.allclose(
                h5_iter_file.root['coordinates'][:], self.dummy_data['iterh5/trajectory'][:]
            ), 'Unable to overwrite trajectory coordinates'

    def test_write_segment_over(self, west_iteration_file):
        """Segment is written successfully in WestIterationFile when there are extra rows."""

        oneframe_data = {'iterh5/trajectory': self.dummy_data['iterh5/trajectory'][:1].copy()}
        self.dummy_data['iterh5/trajectory'][:] *= 2

        segment = Segment(n_iter=3, seg_id=5, data=self.dummy_data)
        segment_oneframe = Segment(n_iter=3, seg_id=5, data=oneframe_data)
        shape = self.dummy_data['iterh5/trajectory'][:].shape

        with WESTIterationFile(self.h5_iter_file_path, mode='a') as h5_iter_file:
            h5_iter_file.write_segment(segment)
            assert np.allclose(h5_iter_file.root['coordinates'][:], self.dummy_data['iterh5/trajectory'][:])
            print(self.dummy_data['iterh5/trajectory'][:])
            print(h5_iter_file.root['coordinates'][:])
            # Rewriting with new coordinates and see if it is written in correctly
            h5_iter_file.write_segment(segment_oneframe)

            assert np.all(h5_iter_file.root['coordinates'][:].shape == shape), 'Extra frames added to per-iter HDF5 File'
            assert np.allclose(
                h5_iter_file.root['coordinates'][:1], oneframe_data['iterh5/trajectory'][:]
            ), 'Unable to overwrite trajectory coordinates'



import logging
import numpy as np
import pickle
import tarfile
from io import BytesIO

import westpa
from westpa.core.trajectory import load_mdtraj, load_netcdf, load_mdanalysis
from westpa.core.h5io import safe_extract

log = logging.getLogger(__name__)


def pcoord_loader(fieldname, pcoord_return_filename, destobj, single_point):
    """Read progress coordinate data into the ``pcoord`` field on ``destobj``.
    An exception will be raised if the data is malformed.  If ``single_point`` is true,
    then only one (N-dimensional) point will be read, otherwise system.pcoord_len points
    will be read.
    """

    system = westpa.rc.get_system_driver()

    assert fieldname == 'pcoord'

    pcoord = np.loadtxt(pcoord_return_filename, dtype=system.pcoord_dtype)

    if single_point:
        expected_shape = (system.pcoord_ndim,)
        if pcoord.ndim == 0:
            pcoord.shape = (1,)
    else:
        expected_shape = (system.pcoord_len, system.pcoord_ndim)
        if pcoord.ndim < 2:
            pcoord.shape = expected_shape
    if pcoord.shape != expected_shape:
        raise ValueError(
            'progress coordinate data has incorrect shape {!r} [expected {!r}] Check pcoord.err or seg_logs for more '
            'information.'.format(pcoord.shape, expected_shape)
        )
    destobj.pcoord = pcoord


def aux_data_loader(fieldname, data_filename, segment, single_point):
    data = np.loadtxt(data_filename)
    segment.data[fieldname] = data
    if data.nbytes == 0:
        raise ValueError('could not read any data for {}'.format(fieldname))


def numpy_data_loader(fieldname, coord_file, segment, single_point):
    log.debug('using numpy_data_loader')
    data = np.load(coord_file, allow_pickle=True)
    segment.data[fieldname] = data
    if data.nbytes == 0:
        raise ValueError('could not read any data for {}'.format(fieldname))


def pickle_data_loader(fieldname, coord_file, segment, single_point):
    log.debug('using pickle_data_loader')
    with open(coord_file, 'rb') as fo:
        data = pickle.load(fo)
    segment.data[fieldname] = data
    if data.nbytes == 0:
        raise ValueError('could not read any data for {}'.format(fieldname))


def mdtraj_trajectory_loader(fieldname, coord_folder, segment, single_point):
    '''Load data from the trajectory return using MDTraj. ``coord_folder`` should be the path to a folder
    containing trajectory files. ``segment`` is the ``Segment`` object that the data is associated with.
    Please see ``load_mdtraj`` for more details. ``single_point`` is not used by this loader.'''
    try:
        data = load_mdtraj(coord_folder)
        segment.data['iterh5/trajectory'] = data
    except Exception as e:
        log.warning('could not read any {} data for HDF5 Framework: {}'.format(fieldname, str(e)))


def netcdf_trajectory_loader(fieldname, coord_folder, segment, single_point):
    '''Load Amber .nc data from the trajectory return. ``coord_folder`` should be the path to a folder
    containing trajectory files. ``segment`` is the ``Segment`` object that the data is associated with.
    Please see ``load_netcdf`` for more details. ``single_point`` is not used by this loader.'''
    try:
        data = load_netcdf(coord_folder)
        segment.data['iterh5/trajectory'] = data
    except Exception as e:
        log.warning('Falling back to default loader for {}: {}'.format(fieldname, str(e)))
        mdtraj_trajectory_loader(fieldname, coord_folder, segment, single_point)


def mdanalysis_trajectory_loader(fieldname, coord_folder, segment, single_point):
    '''Load data from the trajectory return. ``coord_folder`` should be the path to a folder
    containing trajectory files. ``segment`` is the ``Segment`` object that the data is associated with.
    Please see ``load_mdanalysis`` for more details. ``single_point`` is not used by this loader.'''
    try:
        data = load_mdanalysis(coord_folder)
        segment.data['iterh5/trajectory'] = data
    except Exception as e:
        log.warning('Falling back to default loader for {}: {}'.format(fieldname, str(e)))
        mdtraj_trajectory_loader(fieldname, coord_folder, segment, single_point)


def restart_loader(fieldname, restart_folder, segment, single_point):
    '''Load data from the restart return. The loader will tar all files in ``restart_folder``
    and store it in the per-iteration HDF5 file. ``segment`` is the ``Segment`` object that
    the data is associated with. ``single_point`` is not used by this loader.'''
    try:
        with BytesIO() as d:
            with tarfile.open(mode='w:gz', fileobj=d) as t:
                t.add(restart_folder, arcname='.')

            segment.data['iterh5/restart'] = d.getvalue() + b'\x01'  # add tail protection
    except Exception as e:
        log.warning('could not read any {} data for HDF5 Framework: {}'.format(fieldname, str(e)))
    finally:
        d.close()


def restart_writer(path, segment):
    '''Prepare the necessary files from the per-iteration HDF5 file to run ``segment``.'''
    try:
        restart = segment.data.pop('iterh5/restart', None)
        # Making an exception for start states in iteration 1
        if restart is None:
            raise ValueError('restart data is not present')

        with BytesIO(restart[:-1]) as d:  # remove tail protection
            with tarfile.open(fileobj=d, mode='r:gz') as t:
                safe_extract(t, path=path)
    except ValueError as e:
        log.warning('could not write HDF5 Framework restart data for {}: {}'.format(str(segment), str(e)))
        if segment.n_iter == 1:
            log.warning(
                'In iteration 1. Assuming this is a start state and proceeding to skip reading restart from per-iteration HDF5 file for {}'.format(
                    str(segment)
                )
            )
    except Exception as e:
        log.warning('could not write HDF5 Framework restart data for {}: {}'.format(str(segment), str(e)))


def seglog_loader(fieldname, log_file, segment, single_point):
    '''Load data from the log return. The loader will tar all files in ``log_file``
    and store it in the per-iteration HDF5 file. ``segment`` is the ``Segment`` object that
    the data is associated with. ``single_point`` is not used by this loader.'''
    try:
        with BytesIO() as d:
            with tarfile.open(mode='w:gz', fileobj=d) as t:
                t.add(log_file, arcname='seg.log')

            segment.data['iterh5/log'] = d.getvalue() + b'\x01'  # add tail protection
    except Exception as e:
        log.warning('could not read any data for {}: {}'.format(fieldname, str(e)))


def seglog_writer(path, segment):
    '''Untar the log file from segment.'''
    try:
        seglog = segment.data.pop('iterh5/log', None)

        # Raising an exception for start states in iteration 1
        if seglog is None:
            raise ValueError('Log file is not present for segment: {}'.format(str(segment)))

        with BytesIO(seglog[:-1]) as d:  # remove tail protection
            with tarfile.open(fileobj=d, mode='r:gz') as t:
                safe_extract(t, path=path)
    except Exception as e:
        log.warning('could not extract HDF5 Framework log file for {}: {}'.format(str(segment), str(e)))


# Dictionary with all the possible aux dataset loaders
data_loaders = {
    'default': aux_data_loader,
    'auxdata_loader': aux_data_loader,
    'aux_data_loader': aux_data_loader,
    'numpy_loader': numpy_data_loader,
    'npy_loader': numpy_data_loader,
    'numpy_data_loader': numpy_data_loader,
    'npy_data_loader': numpy_data_loader,
    'pickle_loader': pickle_data_loader,
    'pickle_data_loader': pickle_data_loader,
}


# Dictionary with all the possible trajectory loaders
trajectory_loaders = {
    'default': mdtraj_trajectory_loader,
    'trajectory_loader': mdtraj_trajectory_loader,
    'MDTraj_trajectory_loader': mdtraj_trajectory_loader,
    'mdtraj_trajectory_loader': mdtraj_trajectory_loader,
    'amber_trajectory_loader': netcdf_trajectory_loader,
    'netcdf_trajectory_loader': netcdf_trajectory_loader,
    'mda_trajectory_loader': mdanalysis_trajectory_loader,
    'MDAnalysis_trajectory_loader': mdanalysis_trajectory_loader,
    'mdanalysis_trajectory_loader': mdanalysis_trajectory_loader,
}



import sys

from mdtraj import Trajectory
from numpy.testing import assert_allclose
import pytest

from westpa.core.segment import Segment
from westpa.core.trajectory import WESTTrajectory, load_mdanalysis, load_mdtraj, load_netcdf
from westpa.core.propagators.loaders import mdtraj_trajectory_loader, mdanalysis_trajectory_loader, netcdf_trajectory_loader


class TestHDF5Framework:
    """Class to test HDF5 Framework"""

    @pytest.mark.filterwarnings("ignore:Element information is missing")
    def test_load_mdanalysis(self, traj_setup):
        pytest.importorskip('MDAnalysis')

        test_traj = load_mdanalysis(self.current_path)

        assert isinstance(test_traj, WESTTrajectory)

        # load_mdanalysis automatically converts to nm
        assert_allclose(test_traj.xyz / 10, self.ref_coords)
        assert_allclose(test_traj.time, self.ref_time)

    def test_load_netcdf(self, traj_setup):
        test_traj = load_netcdf(self.current_path)

        assert isinstance(test_traj, WESTTrajectory)

        assert_allclose(test_traj.xyz, self.ref_coords)
        assert_allclose(test_traj.time, self.ref_time)

    def test_load_mdtraj(self, traj_setup):
        test_traj = load_mdtraj(self.current_path)

        assert isinstance(test_traj, Trajectory)

        assert_allclose(test_traj.xyz, self.ref_coords)
        assert_allclose(test_traj.time, self.ref_time)

    @pytest.mark.filterwarnings("ignore:Element information is missing")
    def test_mdanalysis_trajectory_loader(self, traj_setup, monkeypatch):
        pytest.importorskip('MDAnalysis')

        dummy_segment = Segment()

        mdanalysis_trajectory_loader('dummy', self.current_path, dummy_segment, False)

        assert_allclose(dummy_segment.data['iterh5/trajectory'].xyz / 10, self.ref_coords)
        assert_allclose(dummy_segment.data['iterh5/trajectory'].time, self.ref_time)

    def test_mdanalysis_trajectory_loader_fail_import(self, traj_setup, monkeypatch):
        """Test fallback to MDTraj with `mdanalysis_trajectory_loader`"""

        dummy_segment = Segment()

        with monkeypatch.context() as m:
            m.setitem(sys.modules, 'MDAnalysis', None)
            mdanalysis_trajectory_loader('dummy', self.current_path, dummy_segment, False)

        assert_allclose(dummy_segment.data['iterh5/trajectory'].xyz, self.ref_coords)
        assert_allclose(dummy_segment.data['iterh5/trajectory'].time, self.ref_time)

    def test_netcdf_trajectory_loader(self, traj_setup):
        dummy_segment = Segment()
        netcdf_trajectory_loader('dummy', self.current_path, dummy_segment, False)

        assert_allclose(dummy_segment.data['iterh5/trajectory'].xyz, self.ref_coords)
        assert_allclose(dummy_segment.data['iterh5/trajectory'].time, self.ref_time)

    def test_mdtraj_trajectory_loader(self, traj_setup):
        dummy_segment = Segment()
        mdtraj_trajectory_loader('dummy', self.current_path, dummy_segment, False)

        assert_allclose(dummy_segment.data['iterh5/trajectory'].xyz, self.ref_coords)
        assert_allclose(dummy_segment.data['iterh5/trajectory'].time, self.ref_time)
